# CamBank
# coding=utf-8

from camlib import BasicPage
from time import localtime, strftime
import logging
import psycopg2.sql as sql

class CamBank (BasicPage.BasicPage):
 def __init__(self, username, data, crypt=None):
  self.log = logging.getLogger(__name__)
  self.log.debug('New CamBank')
  BasicPage.BasicPage.__init__(self, 'CamBank')
  self.username = username
  self.metas['title'] = 'CamBank'
  self.metas['description'] = 'CamBank, le logiciel de compte personnel basé sur les budgets !'
  self._tree = None
  self._full_name = None
  self._id = None
  self.data = data
  self.crypt = crypt
  self.version = 5
  self.epargnes = None
  self.data.logger = self.log
  self.init_from_prefs()
  self.log.debug('CamBank object initialised')
  self.cache = {}

 def get_version(self):
  return 'cambank-{} API-{} schema-{}'.format('2.3.5', self.version, self.schema_version)

 def init_from_prefs(self):
  self.log.debug('Reading Prefs')
  prefs = { r[0]:r[1] for r in self.data.select_all("select name,value from prefs") }
  if 'crypt' in prefs and prefs['crypt'] == 'true':
   self.sympass = self.crypt
  else:
   self.sympass = None
  del self.crypt
  if 'dbversion' in prefs:
   self.schema_version = int(prefs['dbversion'])
  else:
   self.schema_version = 0
  self.log.debug('Prefs applied')

 def symcr(self, str):
  if self.sympass:
   return "pgp_sym_encrypt(%s,'%s')" % \
    (self.data.escape_string(str), self.sympass)
  else:
   return self.data.escape_string(str)

 def symde(self, col):
  if self.sympass:
   col = "pgp_sym_decrypt({}, '{}')".format(col, self.sympass)
  return "convert_from({}::bytea, 'utf8')".format(col)

 # crypt: boolean: do we want to crypt?
 def change_password(self, cur, new, crypt):
  self.log.info('Changing user\'s password, crypt={}'.format(crypt))
  def transform(col):
   newcol = col
   if self.sympass:
    if self.sympass != cur:
     raise Exception('Given password does not match Session password, please check')
    newcol = "pgp_sym_decrypt({}, '{}')".format(newcol, self.sympass)
   if crypt:
    newcol = "pgp_sym_encrypt(text({}), '{}')".format(newcol, new)
   return "{}=bytea({})".format(col, newcol)

  self.data.begin()
  if crypt is not None or self.sympass is not None:
   self.data.execute("update budgets set " + transform('bnom') + "," + transform('bcomment'))
   self.data.execute("update categories set " + transform('cnom'))
   self.data.execute("update operations set " + transform('olibelle') + ',' + transform('odetail'))
   self.data.execute("delete from prefs where name='crypt'")
   if crypt:
    self.data.execute("insert into prefs (name, value) values ('crypt', 'true')")

  if cur != new:
   self.data.execute("update public.users set hash=encode(digest('%s', 'sha512'), 'hex') where login='%s'" % (new, self.username))

  self.data.commit()
  self.log.info('Password/encryption changed')

 def get_last_import_date(self, pattern):
  return self.data.select_one('select ' \
   'to_char(last_import, \'DD/MM/YYYY\') from last_import_date ' \
   'where pattern=\'%s\'' % (pattern))


 """ gestion des operations """

 def create_operation(self, date, detail, montant=0, libelle='Ajout manuel', commit=True):
  self.log.info('Creating operation in DB')
  cur = self.data.execute('insert into operations ' \
   '(oid, odate, olibelle, odetail, omontant) values ' \
   '(nextval(\'oid_seq\'), %s, %s, %s, %d) returning oid' %
   (self.data.escape_string(date) if date != '' else 'now()',
    self.symcr(libelle),
    self.symcr(detail),
    montant))
  oid = cur.fetchone()[0]
  if commit:
   self.data.commit()
  return int(oid)
  self.log.info('Operation inserted')

 def get_last_operation_date(self):
  return self.data.select_one('select ' \
   'to_char(max(odate)+1, \'DD/MM/YYYY\') from operations')

 # def get_pending_operations(self):
 #  return [ Operation(self, line[0], line[1:]) for line in self.data.select_all("select oid, extract(epoch from odate), %s, %s, omontant, obalance, coalesce(hint_bid, 0), account, loadid from operations o where obalance<>omontant or (omontant=0 and not exists (select aid from actions a where a.oid=o.oid)) order by odate" % (self.symde('olibelle'), self.symde('odetail')))]

 def mask_operation(self, oid):
  self.log.info('Hiding operation from the board')
  op = Operation(self, oid)
  op.set_balance(op.omontant)
  self.log.info('Operation masked')

 def get_operations(self, start_date=None):
  return [ Operation(self, line[0], line[1:]) for line in self.data.select_all("select oid, extract(epoch from odate), %s, %s, omontant, obalance, coalesce(hint_bid, 0), account, loadid from operations o" % (self.symde('olibelle'), self.symde('odetail')))]


 """ gestion des categories """

 def create_category(self, nom, ordre):
  self.log.info('Creating new category: {}'.format(nom))
  self.data.execute('update categories ' \
   'set corder=corder+1 where corder >= %d' % (ordre))
  self.data.execute('insert into categories ' \
   '(cnom, corder) values (%s, %d)' % \
   (self.symcr(nom), ordre))
  self.data.commit()
  self.log.info('Category {} created'.format(nom))

 def get_categories(self):
  cats = []
  for line in self.data.select_all('select cid from categories ' \
    'where cid > 0 order by corder'):
   cats.append(Category(self, line[0]))
  return cats


 """ gestion des budgets """

 def create_budget(self, nom, montant, comment, cid):
  self.log.info('Creating new budget: nom={} montant={}'.format(nom, montant))
  self.data.execute('insert into budgets ' \
   '(bnom, bmontant, cid, bcomment) values (%s, %d, %d, %s)' %
   (self.symcr(nom), montant, cid,
    comment and self.symcr(comment) or 'NULL'))
  self.data.commit()
  self.log.info('Budget created')

 def create_treso(self, nom, comment):
  self.log.info('Creating new Treso: nom={}'.format(nom))
  if self.schema_version >= 2:
   try:
    _ = Category(self, 0)
   except:
    self.data.execute("insert into categories (cid,cnom,corder) values (0,'internal',0)")
   self.data.execute('insert into budgets ' \
    '(bnom, bmontant, cid, bcomment, btresorerie) values ' \
    '({}, {}, {}, {}, {})'.format(self.symcr(nom), 0, 0, self.symcr(comment), 'true'))
   self.data.commit()
  else:
   raise TypeError
  self.log.info('Treso created')

 def get_budgets(self):
  if 'budgets' not in self.cache:
   if self.schema_version >= 2:
    self.cache['budgets'] = [ Budget(self, line[0], line[1:]) for line in self.data.select_all('select bid, %s, bmontant, bjauge, %s, cid, bshow, btresorerie from budgets order by bid' % (self.symde('bnom'), self.symde('bcomment'))) ]
   else:
    self.cache['budgets'] = [ Budget(self, line[0], line[1:]) for line in self.data.select_all('select bid, %s, bmontant, bjauge, %s, cid, bshow from budgets order by bid' % (self.symde('bnom'), self.symde('bcomment'))) ]
  return self.cache['budgets']

 def get_sum_budgets(self):
  return self.data.select_one('select sum(bjauge) from budgets where bshow=true and cid<>0') or 0

 def get_sum_tresos(self):
  return self.data.select_one('select sum(bjauge)*-1 from budgets where bshow=true and cid=0') or 0

 def get_debts(self):
  return self.data.select_one('select sum(bjauge) from budgets where bshow=true and bjauge<0 and cid<>0') or 0

 def get_budgets_full(self):
  if self.schema_version >= 2:
   return self.data.select_all('select c.cid, %s, b.bid, %s, %s, bjauge, bmontant, to_char(coalesce(max(adate), \'1900-01-01\'), \'Le DD/MM/YYYY a HH24:MI:SS\'), btresorerie from budgets as b right join categories as c on (b.cid=c.cid) left outer join actions as a on (b.bid=a.bid) where bshow=true or bshow is null group by c.cid, cnom, b.bid, bnom, bcomment, bjauge, bmontant, corder order by corder, b.bid' % (self.symde('cnom'), self.symde('bnom'), self.symde('bcomment')));
  else:
   return self.data.select_all('select c.cid, %s, b.bid, %s, %s, bjauge, bmontant, to_char(coalesce(max(adate), \'1900-01-01\'), \'Le DD/MM/YYYY a HH24:MI:SS\') from budgets as b right join categories as c on (b.cid=c.cid) left outer join actions as a on (b.bid=a.bid) where bshow=true or bshow is null group by c.cid, cnom, b.bid, bnom, bcomment, bjauge, bmontant, corder order by corder, b.bid' % (self.symde('cnom'), self.symde('bnom'), self.symde('bcomment')));

 def get_budgets_summary(self):
  return self.data.select_all('select c.cid, %s, b.bid, %s, %s, bjauge, bmontant, to_char(coalesce(max(adate), \'1900-01-01\'), \'Le DD/MM/YYYY a HH24:MI:SS\') from budgets as b right join categories as c on (b.cid=c.cid) left outer join actions as a on (b.bid=a.bid) where bshow=true or bshow is null group by c.cid, cnom, b.bid, bnom, bcomment, bjauge, bmontant, corder order by corder, b.bid' % (self.symde('cnom'), self.symde('bnom'), self.symde('bcomment')));

 """ gestion des mensualites """

 def create_salary(self, bid, montant):
  self.log.info('Creating new mensualite')
  self.data.execute('insert into mensualites ' \
    '(bid, mmontant) values (%d, %d)' %
    (bid, montant))
  self.data.commit()
  self.log.info('Budget {} will decrease by {} each month!'.format(bid, montant))

 def get_salary(self):
  return self.data.select_all('select mid,bid,mmontant from mensualites order by mid')

 def delete_salary(self, mid):
  self.log.info('Deleting mensualite {}'.format(mid))
  self.data.execute('delete from mensualites where mid=%d' % (mid))
  self.data.commit()
  self.log.info('Mensualite deleted')

 def dispatch_salary(self, oid):
  self.log.info('Distributing the salary {} to budgets'.format(oid))
  sals = self.get_salary()
  for sal in sals:
   self.create_writing(self.get_action_ticket(), oid, sal[2], sal[1])
  self.log.info('Payments done!')


 """ gestion des epargnes """

 def create_epargne(self, nom, min, suppl, max, order):
  self.data.execute('insert into epargnes ' \
   '(ename, emin_jauge, esuppl_jauge, emax_balance, eorder) values (%s, %d, %d, %d, %d)'
   % (self.symcr(nom), min, suppl, max, order))
  self.data.commit()

 def get_epargnes(self):
  if self.epargnes == None:
   self.epargnes = []
   for e in self.data.select_all('select eid, %s, emin_jauge, ' \
     'esuppl_jauge, ebalance, ejauge, emax_balance, %s, eshow, eorder ' \
     'from epargnes order by eorder' %
     (self.symde('ename'), self.symde('ecomment'))):
    self.epargnes.append(Epargne(self, *e))
  return self.epargnes

 def get_epargne(self, eid):
  e = self.data.select_line('select eid, %s, emin_jauge, ' \
   'esuppl_jauge, ebalance, ejauge, emax_balance, %s, eshow, eorder ' \
   'from epargnes where eid=%d' %
   (self.symde('ename'), self.symde('ecomment'), int(eid)))
  return Epargne(self, *e)

 def hide_epargne(self, eid):
  self.data.execute('update epargnes set eshow=false where eid=%d' \
   % (eid))
  self.data.commit()

 def update_epargne(self, eid, nom, min, suppl, max, order):
  self.data.execute('update epargnes set ename=%s, emin_jauge=%d, esuppl_jauge=%d, emax_balance=%d, eorder=%d where eid=%d' % (self.symcr(nom), min, suppl, max, order, eid))
  self.data.commit()


 """ gestion des actions """

 def create_writing(self, aid, oid, montant, bid, epargne=0):
  self.log.info('Binding an operation {} to a budget {}'.format(oid, bid))
  if type(bid) == type('') and bid == 'auto':
   self.epargne(aid, oid, montant)
  elif type(bid) == type('') and bid[0] == 'e':
   ep = self.get_epargne(bid[1:])
   ep.bank(oid, montant, aid, epargne)
  else:
   bud = Budget(self, int(bid))
   bud.bank(oid, montant, aid)
  self.log.info('Action {} successfully recorded'.format(aid))

 def check_action_ticket(self, aid):
  if aid > self.data.select_one('select greatest(0, max(aid)) from actions'):
   return self.data.select_one('select last_value from aid_seq')
  else:
   return -1

 def get_action_ticket(self):
  return self.data.select_one("select nextval('aid_seq')")

 def epargne(self, aid, oid, montant):
  def find_next():
   (eid, reste) = self.data.select_line('select eid, reste from (select ' \
    'eid, emin_jauge-ejauge as reste from epargnes where emin_jauge-ejauge' \
    ' > 0 and eshow=true order by eorder) as a limit 1') or (None, None)
   if reste and reste > 0:
    return (eid, reste)
   return (0, 0)

  def distrib_epargne(eid, combien, aid, oid):
   self.create_writing(aid, oid, combien, 'e%d' % (eid), 1)

  def dispatch(combien, aid, oid):
   init = combien
   es = self.data.select_all('select eid, round(1.0*esuppl_jauge/' \
    '(select sum(esuppl_jauge) from epargnes ' \
    'where ebalance < emax_balance and eshow=true)), ' \
    'emax_balance from epargnes where ebalance < emax_balance and ' \
    'eshow=true')
   for (eid, ratio, max) in es[0:len(es)-2]:
    allowable = min(init*ratio, max)
    distrib_epargne(eid, allowable, aid, oid)
    aid = self.get_action_ticket()
    combien -= allowable
   distrib_epargne(eid, combien, aid, oid)

  while montant > 0:
   (eid, reste) = find_next()
   if eid == 0:
    break
   distrib_epargne(eid, min(montant, reste), aid, oid)
   montant -= min(montant, reste)
   aid = self.get_action_ticket()
  if montant > 0:
   dispatch(montant, aid, oid)

 """
 treso est negatif, c'est un credit. quand on paye avec 12e :
 TX: insert into operation 0e, budget -12e, treso +12e, commit
 """
 def depense_treso(self, aid, date, tresoid, bid, montant, comment):
  self.log.info('Spending money from Treso {} to Budget {}'.format(tresoid, bid))
  if self.schema_version >= 2:
   bud = Budget(self, int(bid))
   treso = Budget(self, int(tresoid))
   oid = self.create_operation(date, comment, libelle='Depense Treso', commit=False)
   op = Operation(self, oid)
   bud.bank(op.oid, montant, aid, commit=False)
   treso.bank(op.oid, -montant, self.get_action_ticket()) # commit here
  else:
   raise TypeError
  self.log.info('Treso spent')

 def get_tresos(self):
  if self.schema_version >= 2:
   return [ b for b in self.get_budgets() if b.btresorerie ]
  return []


""" dans certains contextes, permet de melanger les budgets et les epargnes """

class Boite:
 def __init__(self, cb, type, id, nom, jauge, comment, show):
  cb.log.debug('New Boite')
  self.cb = cb
  self.type = type
  self.id = id
  self.nom = nom
  self.jauge = jauge
  self.comment = comment
  self.show = show
  cb.log.debug('Boite initialised')

 def get_actions(self):
  actions = []
  for line in self.cb.data.select_all('select aid from actions ' \
    'where bid=%d order by adate' % \
    self.id):
   actions.append(Action(self.cb, line[0]))
  return actions


""" budget : une boite où on peut mettre et retirer de l'argent """

class Budget(Boite):
 def __init__(self, cb, bid, data=None):
  cb.log.debug('New Budget')
  self.bid = bid
  if cb.schema_version >= 2:
   if not data:
    data = cb.data.select_line('select %s, bmontant, bjauge, %s, cid, bshow, btresorerie from budgets where bid=%d' % (cb.symde('bnom'), cb.symde('bcomment'), bid))
   if data:
    (self.bnom, self.bmontant, self.bjauge, self.bcomment, self.cid, self.bshow, self.btresorerie) = data
   else:
    raise TypeError
  else:
   row = cb.data.select_line('select %s, bmontant, bjauge, %s, cid, bshow ' \
    'from budgets where bid=%d' % (cb.symde('bnom'), cb.symde('bcomment'),
    bid))
   if row:
    (self.bnom, self.bmontant, self.bjauge, self.bcomment, self.cid, self.bshow) = row
   else:
    raise TypeError
  Boite.__init__(self, cb, 'budget', bid, self.bnom, self.bjauge, self.bcomment, self.bshow)
  cb.log.debug('Budget initialised')

 def get_max_adate(self):
  return localtime(self.cb.data.select_one('select extract(epoch from max(adate)) from actions ' \
   'where bid=%d and amontant > 0' % self.bid))

 def bank(self, oid, montant, aid, commit=True):
  self.cb.log.info('Spending op {} in bud {}'.format(oid, self.bid))
  self.cb.data.begin()
  aid = self.cb.check_action_ticket(aid)
  if aid > 0:
   op = Operation(self.cb, oid)
   self.cb.data.execute('insert into actions (aid, oid, bid, ' \
    'amontant) values (%d, %d, %d, %d)' % \
    (aid, oid, self.bid, montant))
   op.adjust_balance(montant)
   self.adjust_jauge(montant)
   if commit:
    self.cb.data.commit()
  else:
   self.cb.data.rollback()
  self.cb.log.info('Spent')

 def set_nom(self, nom):
  self.cb.log.info('Setting a new name for budget {}: {}'.format(self.bid, nom))
  self.bnom = nom
  self.cb.data.execute('update budgets set bnom=%s where bid=%d' % \
   (self.cb.symcr(nom), self.bid))
  self.cb.data.commit()
  self.cb.log.info('Name set')

 def set_montant(self, montant):
  self.cb.log.info('Setting a new montant {} for budget {}'.format(montant, self.bid))
  self.bmontant = montant
  self.cb.data.execute('update budgets set bmontant=%d where bid=%d' % \
   (montant, self.bid))
  self.cb.data.commit()
  self.cb.log.info('Montant set')

 def adjust_jauge(self, delta):
  self.cb.log.info('Updating budget {} jauge by {}'.format(self.bid, delta))
  self.bjauge += delta
  self.cb.data.execute('update budgets set bjauge=bjauge+%d where bid=%d' % (delta, self.bid))
  self.cb.data.commit()
  self.cb.log.info('Jauge updated')

 def set_comment(self, comment):
  self.cb.log.info('Changing the comment')
  if comment == '':
   self.comment = None
   self.cb.data.execute('update budgets set bcomment=NULL where bid=%d' % (self.bid))
  else:
   self.comment = comment
   self.cb.data.execute('update budgets set bcomment=%s where bid=%d' \
    % (self.cb.symcr(comment), self.bid))
  self.cb.data.commit()
  self.cb.log.info('Comment changed')

 def set_category(self, cid):
  self.cb.log.info('Changing the category')
  self.cid=cid
  self.cb.data.execute('update budgets set cid=%d where bid=%d' % (cid, self.bid))
  self.cb.data.commit()
  self.cb.log.info('Category changed')

 def delete(self):
  self.cb.log.info('Deleting the budget')
  self.cb.data.execute('update budgets set bshow=false where bid=%d' % (self.bid))
  self.cb.data.commit()
  #try:
  # self.cb.data.execute('delete from budgets where bid=%d' % (self.bid))
  # self.cb.data.commit()
  #except:
  # pass
  self.cb.log.info('Budget {} bshow set to false'.format(self.bid))

class Action:
 def __init__(self, cb, aid):
  cb.log.debug('New action object')
  self.cb = cb
  self.aid = aid
  (self.aid, self.adate, self.oid, self.bid, self.amontant) = \
   self.cb.data.select_line('select aid, extract(epoch from adate), oid, bid, amontant from actions where aid=%d' % (aid))
  self.adate = localtime(self.adate)
  cb.log.debug('Action {} initialised'.format(aid))

 def delete(self, commit = True):
  self.cb.log.info('Deleting action {}'.format(self.aid))
  if commit:
   self.cb.data.begin()

  self.cb.data.execute('update operations set obalance = obalance - ' \
   '(select amontant from actions where aid=%d) ' \
   'where oid=(select oid from actions where aid=%d)' % (self.aid, self.aid))
  self.cb.data.execute('update budgets set bjauge = bjauge - ' \
   '(select amontant from actions where aid=%d) where ' \
   'bid=(select bid from actions where aid=%d)' % (self.aid, self.aid))
  self.cb.data.execute('update epargnes set ebalance = ebalance - ' \
   '(select amontant from actions where aid=%d) where ' \
   'eid=(select bid from actions where aid=%d)' % (self.aid, self.aid))
  self.cb.data.execute('delete from actions where aid=%d' % (self.aid))

  if self.cb.schema_version >= 2 and Budget(self.cb, self.bid).cid == 0:
   for act in Operation(self.cb, self.oid).get_actions():
    act.delete(commit = False)
   self.cb.data.execute('delete from operations where oid=%d' % (self.oid))

  if commit:
   self.cb.data.commit()
  self.cb.log.info('Action {} rollbacked'.format(aid))


class Operation:
 def __init__(self, cb, oid, data=None):
  cb.log.debug('New Operation object for oid={}'.format(oid))
  self.cb=cb
  self.oid = oid
  if not data:
   data = cb.data.select_line('select extract(epoch from odate), %s, ' \
    '%s, omontant, obalance, coalesce(hint_bid, 0), account, loadid from operations where oid=%d' % \
    (self.cb.symde('olibelle'), self.cb.symde('odetail'), oid))
  (self.odate, self.olibelle, self.odetail, self.omontant, self.obalance,
   self.hint_bid, self.account, self.loadid) = data
  self.odate = localtime(self.odate)
  cb.log.debug('Operation initialised')

 def __str__(self):
  return '({}, {}, {}, {})'.format(self.oid, strftime('%Y-%m-%d', self.odate), self.omontant, self.obalance)

 def set_balance(self, balance):
  self.cb.log.info('Setting the balance')
  self.obalance = balance
  self.cb.data.execute('update operations set obalance=%d where oid=%d' % (balance, self.oid))
  self.cb.data.commit()
  self.cb.log.info('Balance set')

 def adjust_balance(self, delta):
  self.cb.log.info('Updating operation {} balance by {}'.format(self.oid, delta))
  self.obalance += delta
  self.cb.data.execute('update operations set obalance=obalance+%d where oid=%d' % \
   (delta, self.oid))
  self.cb.data.commit()
  self.cb.log.info('Balance updated')

 def get_actions(self):
  aids = []
  for line in self.cb.data.select_all('select aid from actions where oid=%d' % (self.oid)):
   aids.append(Action(self.cb, line[0]))
  return aids

 def get_type(self):
  return {
   'CARTE': 'CARTE',
   'PAIEMENT': 'CARTE',
   'RETRAIT': 'RETRAIT',
   'VIREMENT': 'VIREMENT',
   'VIR': 'VIREMENT',
   'PRLV': 'PRELEVEMENT',
   'CHEQUE': 'CHEQUE',
   'CHQ.': 'CHEQUE',
  }.get(self.olibelle.split(' ', 1)[0], 'DIVERS')


class Category:
 def __init__(self, cb, cid):
  cb.log.debug('New category')
  self.cb=cb
  self.cid=cid
  (self.cnom, self.corder) = cb.data.select_line('select %s, corder ' \
   'from categories where cid=%d' % (cb.symde('cnom'), cid))
  cb.log.debug('Category object initialised')

 def set_nom(self, nom):
  self.cb.log.info('Setting a name for a cat')
  if self.cid == 0:
   raise UnboundLocalError("this category has been deleted")
  self.cnom = nom
  self.cb.data.execute('update categories set cnom=%s where cid=%d' % \
   (self.cb.symcr(nom), self.cid))
  self.cb.data.commit()
  self.cb.log.info('Name set')

 def set_order(self, order):
  self.cb.log.info('Setting the order')
  if self.cid == 0:
   raise UnboundLocalError("this category has been deleted")
  if self.corder == order:
   pass
  self.cb.data.execute('update categories set corder=0 where cid=%d' % (self.cid))
  self.cb.data.execute('update categories set corder=corder-1 where corder > %d and corder <= %d' \
   % (self.corder, order))
  self.cb.data.execute('update categories set corder=corder+1 where corder < %d and corder >= %d' \
   % (self.corder, order))
  self.cb.data.execute('update categories set corder=%d where cid=%d' % (order, self.cid))
  self.corder = order
  self.cb.data.commit()
  self.cb.log.info('Order set')

 def get_budgets(self):
  #if self.cid == 0:
  # raise UnboundLocalError("this category has been deleted")
  buds = []
  for line in self.cb.data.select_all('select bid from budgets where ' \
    'cid=%d order by bid' % self.cid):
   buds.append(Budget(self.cb, line[0]))
  return buds

 def delete(self):
  self.cb.log.info('Deleting category {}'.format(self.cid))
  if self.cid == 0:
   pass
  # FIXME: create a Error handling mecanism and use it here and elsewhere
  #try:
  # self.cb.data.execute('delete from categories where cid=%d' % (self.cid))
  # self.cb.data.commit()
  # self.cid = 0
  # self.cnom = "deleted"
  #except:
  # pass
  self.cb.data.execute('delete from categories where cid=%d' % (self.cid))
  self.cb.data.commit()
  self.cid = 0
  self.cnom = "deleted"
  self.cb.log.info('Category deleted')

class Epargne(Boite):
 def __init__(self, cb, eid, ename, emin_jauge, esuppl_jauge, ebalance, ejauge, emax_balance, ecomment, eshow, eorder):
  Boite.__init__(self, cb, 'epargne', eid, ename, ebalance, ecomment, eshow)
  self.eid = eid
  self.ename = ename
  self.emin_jauge = emin_jauge
  self.esuppl_jauge = esuppl_jauge
  self.ebalance = ebalance
  self.ejauge = ejauge
  self.emax_balance = emax_balance
  self.ecomment = ecomment or ""
  self.eshow = eshow
  self.eorder = eorder

 def bank(self, oid, montant, aid, epargne=0):
  self.cb.data.begin()
  aid = self.cb.check_action_ticket(aid)
  if aid > 0:
   op = Operation(self.cb, oid)
   self.cb.data.execute('insert into actions (aid, oid, bid, ' \
    'amontant) values (%d, %d, %d, %d)' % \
    (aid, oid, self.eid, montant))
   self.ebalance += montant
   self.cb.data.execute('update epargnes set ebalance=ebalance+%d ' \
    'where eid=%d' % (montant, self.eid))
   op.adjust_balance(montant)
   if epargne:
    self.cb.data.execute('update epargnes set ejauge=ejauge+%d ' \
     'where eid=%d' % (montant, self.eid))
    self.ejauge += montant
   self.cb.data.commit()
  else:
   self.cb.data.rollback()



# coding=utf-8
from __future__ import print_function
import sys, datetime, re

class BankItem:
 def __init__(self):
  self.date = None
  self.amount = None
  self.cleared = None
  self.num = None
  self.payee = None
  self.memo = None
  self.address = None
  self.category = None
  self.categoryInSplit = None
  self.memoInSplit = None
  self.amountOfSplit = None
  self.checkNumber = None

 def getCamBankCSV(self):
  return "%s;%s;%s;%s;%s" % (self.date, self.payee or self.memo, self.memo or self.payee, self.amount, 'EUR')

class QifItem(BankItem):
 pass

def parseQif(infile):
 """
 Parse a qif file and return a list of entries.
 infile should be open file-like object (supporting readline() ).
 """

 inItem = False
 englishDateFormat = False

 items = []
 curItem = QifItem()

 content = infile.read()
 if 'decode' in dir(content):
  content = content.decode()
 for line in re.split('\n|\r|\n\r', content)[1:]:
  if line == '': # blank line
   pass
  elif line[0] == '^': # end of item
   # save the item
   items.append(curItem)
   curItem = QifItem()
  elif line[0] == 'D':
   curItem.date = line[1:]
   # try to detect english Date Format if month > 12 or if date is in the future
   (d, m, y) = curItem.date.split('/')
   if int(m) > 12 or datetime.date(int(y), int(m), int(d)) > datetime.date.today():
    englishDateFormat = True
   # correct YY to YYYY
   if int(y) < 2000:
    curItem.date = d+'/'+m+'/20'+y
  elif line[0] == 'T':
   curItem.amount = line[1:]
  elif line[0] == 'C':
   curItem.cleared = line[1:]
  elif line[0] == 'P':
   curItem.payee = line[1:]
  elif line[0] == 'N':
   curItem.checkNumber = line[1:]
  elif line[0] == 'M':
   curItem.memo = line[1:]
  elif line[0] == 'A':
   curItem.address = line[1:]
  elif line[0] == 'L':
   curItem.category = line[1:]
  elif line[0] == 'S':
   try:
    curItem.categoryInSplit.append(";" + line[1:])
   except AttributeError:
    curItem.categoryInSplit = line[1:]
  elif line[0] == 'E':
   try:
    curItem.memoInSplit.append(";" + line[1:])
   except AttributeError:
    curItem.memoInSplit = line[1:]
  elif line[0] == '$':
   try:
    curItem.amountInSplit.append(";" + line[1:])
   except AttributeError:
    curItem.amountInSplit = line[1:]
  else:
   # don't recognise this line; ignore it
   print("Skipping unknown line:\n{}".format(line), file=sys.stderr)

  line = infile.readline()
  if "decode" in dir(line):
   line = line.decode()

 # reverting to european dates if needed
 if englishDateFormat:
  ii = []
  for item in items:
   (m, d, y) = item.date.split('/')
   item.date = d+'/'+m+'/'+y
   ii.append(item)
  return ii
 return items





class OfxItem(BankItem):
 pass

def parseOfx(infile):
 """
 Parse an OFXv1 file and return a list of entries.
 infile should be open file-like object (supporting readline() ).
 """
 items = []
 from ofxtools import OFXTree
 tree = OFXTree()
 tree.parse(infile)
 response = tree.convert()
 for st in response.statements:
  for tr in st.transactions:
   curItem = OfxItem()
   curItem.date = tr.dtposted.strftime('%d/%m/%Y')
   curItem.amount = str(tr.trnamt)
   curItem.payee = tr.name
   curItem.memo = tr.memo
   # save the item
   items.append(curItem)
 return items



def parseEpaymentsCSV(infile):
 """
 parse a CSV file downloaded at epayments.com. returns the typical list of Bankitem
 """
 items = []
 content = infile.read()
 if 'decode' in dir(content):
  content = content.decode()
 for line in re.split('\n|\r|\n\r', content)[1:]:
  if line == '':
   continue
  row = line.split(';')
  curItem = BankItem()
  curItem.date = datetime.datetime.strptime(row[0].split(' ')[0], '%m/%d/%Y').strftime('%d/%m/%Y')
  curItem.amount = row[4]
  curItem.payee = 'CARTE ' + re.compile('  *').sub(' ', row[3].replace('\\', ' '))
  curItem.memo = curItem.payee
  items.append(curItem)
 return items


if __name__ == "__main__":
 # open and read arg1 and write CSV to stdout
 with open(sys.argv[1]) as f:
  items = parseEpaymentsCSV(f)
  #items = parseOfx(f)
  #print repr(items[0])
  for item in items:
   print(item.getCamBankCSV())



# coding=utf-8
from camlib import InputController
from cambank import CamBank, CamBankHTML

class Input(InputController.InputController):
 def __init__(self, page, form):
  self._page = page
  self._form = form

 def process_input(self):
  page = self._page
  form = self._form

  if 'tresoid' in form and 'aid' in form and 'montant' in form and 'bid' in form:
   tresoid = int(form['tresoid'].value)
   if 'label' in form:
    label = form['label'].value
   else:
    treso = CamBank.Budget(page, tresoid)
    label = 'Depense de tresorerie {}'.format(treso.bnom)
   page.depense_treso(int(form['aid'].value), '', tresoid,
     int(form['bid'].value), float(form['montant'].value.replace(',', '.'))*-100,
     label)
   return "index.cgi"

  if 'setpass' in form and 'curpass' in form and \
    'newpass' in form and 'crypt' in form:
   page.change_password(form['curpass'].value, form['newpass'].value, form['crypt'].value == 'y' and True or False)
   return "index.cgi"

  if 'deloid' in form:
   page.mask_operation(int(form['deloid'].value))
   return "index.cgi"

  if 'dispatchsalary' in form:
   page.dispatch_salary(int(form['dispatchsalary'].value))
   return "index.cgi"

  if 'aid' in form and 'oid' in form and 'montant' in form \
    and 'bid' in form:
   page.create_writing(int(form['aid'].value), int(form['oid'].value),
    int("%.0f" % (float(form['montant'].value.replace(',','.'))*100)),
    form['bid'].value)
   return "index.cgi"

  if 'date' in form and 'detail' in form \
    and 'montant' in form:
   page.create_operation(form['date'].value, form['detail'].value,
    int(float(form['montant'].value)*100))
   return "index.cgi"

  if 'newcnom' in form and 'corder' in form:
   page.create_category(form['newcnom'].value, int(form['corder'].value))
   return "index.cgi?editcat=1"

  if 'cid' in form and 'cnom' in form and 'corder' in form:
   cat = CamBank.Category(page, int(form['cid'].value))
   cat.set_nom(form['cnom'].value)
   cat.set_order(int(form['corder'].value))
   return "index.cgi?editcat=1"

  if 'delcatid' in form:
   cat = CamBank.Category(page, int(form['delcatid'].value))
   cat.delete()
   return "index.cgi?editcat=1"

  if 'newbnom' in form and 'cid' in form: # and 'bmontant' in form:
   if form['cid'].value == '0':
    page.create_treso(form['newbnom'].value, form['bcomment'].value if 'bcomment' in form else '')
   elif 'bmontant' in form:
    page.create_budget(form['newbnom'].value, int(float(form['bmontant'].value)*100),
    'bcomment' in form and form['bcomment'].value or '', int(form['cid'].value))
   return "index.cgi?editcatid=%d" % (int(form['cid'].value))

  if 'bnom' in form and 'cid' in form and 'bid' in form and 'bmontant' in form:
   bud = CamBank.Budget(page, int(form['bid'].value))
   cid = bud.cid+0
   bud.set_nom(form['bnom'].value)
   bud.set_category(int(form['cid'].value))
   bud.set_comment('bcomment' in form and form['bcomment'].value or '')
   bud.set_montant(int(float(form['bmontant'].value)*100))
   return "index.cgi?editcatid=%d" % (cid)

  if 'delbid' in form:
   bud = CamBank.Budget(page, int(form['delbid'].value))
   cid = bud.cid
   bud.delete()
   return "index.cgi?editcatid=%d" % (cid)

  if 'cancelaid' in form:
   act = CamBank.Action(page, int(form['cancelaid'].value))
   act.delete()
   if 'histobid' in form:
    return "index.cgi?histobid=%d" % (int(form['histobid'].value))
   else:
    return "index.cgi"

  if 'salary' in form and 'bid' in form and 'montant' in form:
   page.create_salary(int(form['bid'].value), int(float(form['montant'].value)*100))
   return "index.cgi?salary=1"

  if 'salary' in form and 'mid' in form:
   page.delete_salary(int(form['mid'].value))
   return "index.cgi?salary=1"

  if 'epargne' in form and 'nom' in form and 'min' in form \
    and 'max' in form and 'suppl' in form and 'order' in form:
   page.create_epargne(form['nom'].value, int(form['min'].value)*100,
    int(form['suppl'].value)*100, int(form['max'].value)*100,
    int(form['order'].value))
   return "index.cgi?epargne=1"

  if 'hideeid' in form:
   page.hide_epargne(int(form['hideeid'].value))
   return "index.cgi?epargne=1"

  if 'seteid' in form and 'nom' in form and 'min' in form \
    and 'max' in form and 'suppl' in form and 'order' in form:
   page.update_epargne(int(form['seteid'].value), form['nom'].value,
    int(form['min'].value)*100, int(form['suppl'].value)*100,
    int(form['max'].value)*100, int(form['order'].value))
   return "index.cgi?epargne=1"



# CamBankHTML
# coding=utf-8

import locale
from camlib import BasicHTMLController
from cambank.CamBank import *
from cambank import QifParser
from cambank.utils import filter_pending_operations
from time import strftime
import re

class HTML(BasicHTMLController.BasicHTMLController):
 def __init__(self, parent):
  BasicHTMLController.BasicHTMLController.__init__(self, parent)
  locale.setlocale(locale.LC_ALL, ('fr_FR', 'utf-8'))
  self.html_version = 2

 def get_headers(self):
  html = BasicHTMLController.BasicHTMLController.get_headers(self)
  return '''
<meta name="application-name" content="CamBank"/>
<link rel="shortcut icon" href="../images/cambank-16x16.png">
<link rel="icon" href="../images/cambank-32x32.png" sizes="32x32"/>
<link rel="icon" href="../images/cambank-48x48.png" sizes="48x48"/>''' + html

 def show_headers(self):
  return """
<div style="text-align: center; margin-top: 0" class="box">
<a href="index.cgi"><img src="../images/cambank.png" alt="CamBank Logo"/></a>
<h6 style="margin-top: 0">Le logiciel de compte personnel<br/>bas&eacute; sur les budgets&nbsp;!</h6>
<h4>
<a href="index.cgi?prefs=1">{}</a>
(<a href="index.cgi?logout=1">D&eacute;connecter</a>)
</h4>
<h4>
<a href="index.cgi?import=0">Importer des &eacute;critures</a> &mdash;
<!-- <a href="index.cgi?epargne=1">&Eacute;pargnes</a> &mdash; -->
<a href="index.cgi?salary=1">Mensualit&eacute;s</a> &mdash;
<a href="index.cgi?help=1">Aide</a>
</h4>
</div>
""".format(self._parent.username)

 def get_style(self):
  style = '''
body {
 font-family: sans-serif, Arial;
 font-size: smaller;
 background-color: #EBFFEC;
}
.edit {
 font-size: smaller;
}
h2 {
 font-weight: bold;
 color: white;
 padding: 0.2em;
 padding-left: 0.8em;
 margin: 1px;
 width: 50%;
 background-color: #3C6C42;
}
input, select {
 font-family: sans-serif, Arial;
 background-color: #D4FFD9;
 border: solid 1px;
}
.box > table {
 background-color: #3C6C42;
 padding: 1px;
 margin: 0;
 margin-left: 5em;
}
.box > table td {
 padding: 0.3em 1em;
 background-color: #D4FFD9;
}
.box > table th {
 color: white;
}
.detail {
 width: 20em;
}
img {
 border: 0;
}
.box {
 margin-top: 5em;
}
.category_name {
 font-size: larger;
}
a {
 color: black;
 font-weight: bold;
/* text-decoration: none; */
}
td.number {
 text-align: right;
}
'''
  return style

 def get_cached_budgets(self):
  if 'cached_budgets' not in dir(self):
   self.cached_objects = self.cb.get_budgets_full()
  return self.cached_budgets

 def show_retour_accueil(self):
  return """
<p>
<a href="index.cgi">Revenir à la page d'accueil</a>
</p>
"""

 def show_footers(self):
  cb = self._parent
  return """
<h6 id="footer">
CamBank &copy; 2008-2018 Camille Huot.<br/>
{} HTML-{}
</h6>
""".format(cb.get_version(), self.html_version)

class Home(HTML):
 def __init__(self, cb):
  HTML.__init__(self, cb)

 def get_body(self):
  html = ''
  html += self.show_headers()
  html += self.show_pending_operations()
  html += self.show_tresos()
  html += self.show_budgets()
  #html += self.show_epargnes()
  html += self.show_create_operation_form()
  html += self.show_footers()
  return html

 def show_create_operation_form(self):
  return '''
<div class="box">
<h2>Ajouter une op&eacute;ration manuelle</h2>
<form>
<table>
<tr><th align="right">Date :</th><td><input type="text" name="date"/></td></tr>
<tr><th align="right">Details :</th><td><input type="text" name="detail"/></td></tr>
<tr><th align="right">Montant :</th><td><input type="text" name="montant"/></td></tr>
<tr><th></th><td><input type="image" src="../images/valid.png" style="border: 0"/></td></tr>
</table>
</form>
</div>'''

 def print_budgets_menu(self, default = 0, exclude = -1):
  cb = self._parent
  buds = cb.get_budgets()
  budgets_item = "<select name=\"bid\">\n"
  for bud in buds:
   if bud.bshow is False:
    continue
   if bud.bid == exclude:
    continue
   if bud.bid == default:
    budgets_item += "<option value=\"%d\" selected>%s</option>\n" % \
     (bud.bid, bud.bnom)
   else:
    budgets_item += "<option value=\"%d\">%s</option>\n" % \
     (bud.bid, bud.bnom)
  budgets_item += "</select>\n"
  return budgets_item

 def show_pending_operations(self):
  cb = self._parent
  actionid = cb.get_action_ticket()
  ops = cb.get_operations()
  if len(ops) == 0:
   return ''

  # montant minimum pour considerer l'operation comme un salaire
  salaire = 0
  for sal in cb.get_salary():
   salaire += sal[2]

  ops = filter_pending_operations(ops)

  # remplissage du template
  html = '<div class="box">'
  for optype in sorted(set([ o.get_type() for o in ops ])):
   html += "<h2>Op&eacute;rations en attente de traitement : %s</h2>\n" % (optype)
   html += "<table>\n<tr>\n"
   html += "<th>Date</th>\n"
   html += "<th>D&eacute;tails de l'op&eacute;ration</th>\n"
   html += "<th>Euros</th>\n"
   html += "<th>Allouer cette op&eacute;ration &agrave; un budget</th>\n"
   html += "<th>Supprimer</th>\n"
   html += "</tr>\n"
   myops = [ o for o in ops if o.get_type() == optype ]
   nbops=len(myops)
   for op in myops:
     salairehtml = ''
     if salaire > 0 and op.omontant-op.obalance > salaire:
      salairehtml += '<tr><td>Traiter comme un salaire (allouer automatiquement %d EUR dans les budgets predefinis)&nbsp;:</td><td>' % (salaire/100)
      salairehtml += '<a href="index.cgi?dispatchsalary=%d"><img src="../images/valid.png" style="border: 0"/></a>' % (op.oid)
      salairehtml += '</td><tr><tr><td colspan="2" style="text-align: center"><hr/></td></tr>'

     html += """
     <tr>
     <td class="date">
     %s
     </td>
     <td class="detail">
     %s
     </td>
     <td>
     %.2f
     </td>
     <td>
     <form>
     <input type="hidden" name="oid" value="%d"/>
     <input type="hidden" name="aid" value="%d"/>
     <table>%s
     <tr><td align="right">Allouer&nbsp;:</td><td>
     <input name="montant" value="%.2f" size="4" style="text-align: right;"/>&thinsp;E</td></tr>
     <tr><td align="right">Au budget&nbsp;:</td><td>
     """ % (
       strftime('%Y-%m-%d', op.odate),
       'similar' in op.__dict__ and '{} <hr/><div>Similaires:<ul>{}</ul></div>'.format(op.odetail, "\n".join(["<li>{} [account={} loadid={}]</li>".format(op2.odetail, op2.account, op2.loadid) for op2 in op.similar])) or op.odetail,
       float(op.omontant)/100,
       op.oid, actionid,
       salairehtml,
       float((op.omontant-op.obalance))/100
     )
     html += self.print_budgets_menu(op.hint_bid)
     html += """
     </tr>
     <tr><td></td><td><input type="image" src="../images/valid.png" style="border: 0"/></td>
     </table>
     </form>
     </td>
     <td>
     <a href="index.cgi?deloid=%d"><img src="../images/delete.png" title="Supprimer cette op&eacute;ration"/></a>
     </td>
     </tr>
     """ % (op.oid)
   html += "<tr><th colspan=\"5\">%d sur %d op&eacute;ration(s) en attente de traitement</th></tr></table>" % (len(myops), len(ops))
  html += "</div>\n"
  return html

 def show_tresos(self):
  cb = self._parent
  tresos = cb.get_tresos()
  actionid = cb.get_action_ticket()
  if len(tresos) == 0:
   return ""
  html = """
<div class="box">
<h2>Trésorerie</h2>
<table>
<tr><th><span class="category_name">Nom</span></th><th>Montant</th><th>Créer opération</th></tr>
"""
  for treso in tresos:
   if treso.bcomment:
    treso.bcomment = '<br/>(%s)' % (treso.bcomment.replace("\n", '<br/>'))
   else:
    treso.bcomment = ''
   html += '<tr><td><a href="index.cgi?histobid=%d">%s</a>%s</td>' % (treso.bid, treso.bnom, treso.bcomment)
   html += locale.format_string('<td class="number">%.2f</td>', float(-1*treso.bjauge)/100)
   html += '''<td><form>Dépense
     <input type="hidden" name="tresoid" value="{}"/>
     <input type="hidden" name="aid" value="{}"/>
     <input name="montant" size="4" style="text-align: right;"/>&thinsp;E
     pour '''.format(treso.bid, actionid)
   html += self.print_budgets_menu(exclude = treso.bid)
   html += """ - label <input name="label" size="20" style="text-align: right;"/>
     <input type="image" src="../images/valid.png" style="border: 0"/>
     </form></td></tr>"""
  html += '''
<tr><th colspan="3" class="edit"><a href="index.cgi?editcatid=0">(éditer les trésoreries)</a>
</th></tr>
</table>
<p>Total des tresos&nbsp;: %.2f Euros.</p>
</div>''' % (float(cb.get_sum_tresos())/100)
  return html

 def show_budgets(self):
  cb = self._parent
  html = """
<div class="box">
<h2>Liste des budgets</h2>
<table>
"""
  cur_cat = -1
  for (cid, cnom, bid, bnom, bcomment, bjauge, bmontant, blastdate) in cb.get_budgets_summary():
   if cid == 0:
    continue
   if cid != cur_cat:
    cur_cat = cid
    html += '<tr><th><span class="category_name">Categorie %s</span><br/><span class="edit">' \
     '<a href="index.cgi?editcatid=%d">(&eacute;diter les budgets)</a></span></th>' \
     % (cnom, cid)
    if bid is None:
     html += '<th solspan="3"></th></tr>'
     continue
    html += '<th>Montant<br/>allou&eacute;</th>' \
     '<th>Montant<br/>cible</th><th>Date du dernier d&eacute;p&ocirc;t</th></tr>'
   if bcomment:
    bcomment = '<br/>(%s)' % (bcomment.replace("\n", '<br/>'))
   else:
    bcomment = ''
   html += '<tr><td><a href="index.cgi?histobid=%d">%s</a>%s</td>' % \
    (bid, bnom, bcomment)
   html += locale.format_string('<td class="number">%.2f</td>', float(bjauge)/100)
   html += locale.format_string('<td class="number">%d</td>', float(bmontant)/100)
   html += '<td style="text-align: center">%s</td></tr>' % (blastdate and blastdate or 'Jamais')
  html += '<tr><th colspan="4" class="edit"><a href="index.cgi?editcat=1">(&eacute;diter les ' \
   "cat&eacute;gories)</a></th></tr>\n"
  html += "</table>\n"
  html += "<p>Total des budgets&nbsp;: %.2f Euros.</p>" % (float(cb.get_sum_budgets())/100)
  html += "<p>Total des dettes&nbsp;: %.2f Euros.</p>" % (float(cb.get_debts())/100)
  html += "</div>"
  return html

 def show_epargnes(self):
  cb = self._parent
  html = """
<div class="box">
<h2>Épargnes</h2>
<table>
<tr>
<th>Épargne</th>
<th>Montant<br/>allou&eacute;</th>
<th>Montant<br/>cible</th>
<th>Montant<br/>réel</th>
</tr>
"""
  for ep in cb.get_epargnes():
   if ep.eshow is False:
    continue
   if ep.ecomment:
    comment='<br/>(%s)' % (ep.ecomment.replace("\n", '<br>'))
   else:
    comment=''
   html += """
<tr>
<td><a href="index.cgi?histobid=%d">%s</a>%s</td>
<td class="number">%s</td>
<td class="number">%s</td>
<td class="number">%s</td>
</tr>
""" % (ep.eid, ep.ename, comment, locale.format_string('%.2f',
    float(ep.ejauge)/100), locale.format_string('%d',
    float(ep.emax_balance)/100),
    locale.format_string('%.2f', float(ep.ebalance)/100))
  html += """
</table>
"""
  return html


class Histo(HTML):
 def __init__(self, cb, bid):
  HTML.__init__(self, cb)
  self.bid = bid

 def get_body(self):
  html = ''
  html += self.show_headers()
  html += self.show_budget_history()
  html += self.show_retour_accueil()
  html += self.show_footers()
  return html

 def show_budget_history(self):
  not_a_budget = 0
  try:
   bud = Budget(self._parent, self.bid)
  except:
   not_a_budget = 1

  if not_a_budget:
   bud = self._parent.get_epargne(self.bid)
   if not bud:
    return 'nonono'

  html = ''
  html += """
<div class="box">
<h2>Historique de la boite %s</h2>
<table>
<tr><th>Date action</th><th>D&eacute;tails</th><th>Date op&eacute;ration</th><th>Montant</th><th>Annuler</th></tr>
""" % (bud.nom)
  for action in reversed(bud.get_actions()):
   op = Operation(self._parent, action.oid)
   html += """
<tr><td>%s</td><td><a href="index.cgi?viewoid=%d">%s</a></td><td align="right">%s</td><td align="right">%.2f</td>
 <td><a href="index.cgi?histobid=%d&amp;cancelaid=%d"><img src="../images/undo.png" alt="Annuler cette action" /></a></td></tr>
""" % (strftime('Le %Y-%m-%d &agrave; %H:%M:%S', action.adate), op.oid, op.odetail, strftime('%Y-%m-%d', op.odate), float(action.amontant)/100, self.bid, action.aid)
  html += """
<tr><th colspan="4" align="right">Montant actuel du budget</th><th>%.2f</th></tr>
</table>
</div>
""" % (float(bud.jauge)/100)
  return html

class UploadImport(HTML):
 """
 Cette page gere toute la partie IMPORTATION des ecritures bancaires dans CamBank.
 - Un texte explique brievement ce qu'il faut faire
 - Un formulaire permet d'uploader un fichier QIF
 - Une boite affiche le resultat de la lecture de ce fichier
 - Un tableau permet de selectionner quelles ecritures inserer
 - Une boite affiche le resultat de l'insertion
 """

 def __init__(self, cb):
  HTML.__init__(self, cb)
  self.file = 'NO'
  self.oids = []

 def get_body(self):
  html = ''
  html += self.show_headers()
  html += self.show_upload_help()
  if self.file != 'NO':
   html += self.read_csv()
  html += self.load_operations()
  html += self.show_upload_form()
  html += self.show_retour_accueil()
  html += self.show_footers()
  return html

 def show_upload_help(self):
  html = """
<p>
Cette page permet d'inserer dans CamBank des ecritures provenant de votre compte
bancaire.
</p>
<p>
Pour cela, vous devez telecharger depuis votre interface bancaire un export au format QIF (Money) et le chargez ici.
</p>
<p>
Vous aurez ensuite la possibilite d'inserer, ligne par ligne, les ecritures dans CamBank ou eventuellement de les supprimer.
</p>
<p>
Vous pouvez supprimer, par exemple, les transferts entre comptes, qui sont donc ni des entrees ni des sorties d'argent.
</p>
<h4>Quelques liens pour telecharger des ecritures</h4>
<table>
<tr>
<td><img src="../images/logo_ingdirect.gif"/></td>
<td><a href="https://secure.ingdirect.fr/public/displayLogin.jsf">ING Direct</a></td>
<td>Chargez le releve a partir du : %s</td>
</tr>
<tr>
<td><img src="../images/logo_boursorama.gif"/></td>
<td><a href="https://clients.boursorama.com/documents/generer">Boursorama</a></td>
<td>Chargez le releve a partir du : %s</td>
</tr>
<tr>
<td><img src="../images/logo_bnp.png"/></td>
<td><a href="https://mabanque.bnpparibas/fr/connexion/virements-services/telechargement-des-operations">BNP Paribas COURANT</a></td>
<td>Chargez le releve a partir du : %s</td>
</tr>
<tr>
<td><img src="../images/logo_bnp.png"/></td>
<td><a href="https://mabanque.bnpparibas/fr/connexion/virements-services/telechargement-des-operations">BNP Paribas COMMUN</a></td>
<td>Chargez le releve a partir du : %s</td>
</tr>
<tr>
<td><img src="../images/logo_sg.gif"/></td>
<td><a href="https://www.particuliers.societegenerale.fr/">Soci&eacute;t&eacute;G&eacute;n&eacute;rale</a></td>
<td>Chargez le releve a partir du : %s</td>
</tr>
<tr>
<td><img src="../images/logo-fortuneo.gif"/></td>
<td><a href="https://www.fortuneo.fr/fr/identification.jsp">Fortuneo</a></td>
<td>Chargez le releve a partir du : %s</td>
</tr>
<tr>
<td><img src="../images/logo_lcl.gif"/></td>
<td><a href="https://particuliers.secure.lcl.fr/index.html">Cr&eacute;dit Lyonnais</a></td>
<td>Chargez le releve a partir du : %s</td>
</tr>
</table>
""" % (
    self._parent.get_last_import_date('40000441192_[0-9_]*.qif') or '[aucune op&eacute;ration dans la base]',
    self._parent.get_last_import_date('00040094318_R[0-9]*.[0-9]*.qif') or '[aucune op&eacute;ration dans la base]',
    self._parent.get_last_import_date('E2900888.*qif') or '[aucune op&eacute;ration dans la base]',
    self._parent.get_last_import_date('E2903982.*qif') or '[aucune op&eacute;ration dans la base]',
    self._parent.get_last_import_date('00057693757.*.qif') or '[aucune op&eacute;ration dans la base]',
    self._parent.get_last_import_date('HistoriqueOperations_.*.qif') or '[aucune op&eacute;ration dans la base]',
    self._parent.get_last_import_date('T_cpte_.*.qif') or '[aucune op&eacute;ration dans la base]',
   )
  return html

 def read_csv(self):
  """
  Ouvre le fichier envoye par le client et extrait les lignes.
  Le fichier est actuellement traite obligatoirement comme un fichier QIF.
  Il est passe a un convertisseur CSV pour reutiliser le code existant.
  """
  db = self._parent.data
  html = ""
  html += """
<div class="box">
<h2>Lecture du fichier CSV</h2>
<p>Ouverture du fichier&nbsp;:
"""
  try:
   if self.file.filename.lower().endswith('.ofx'):
    items = QifParser.parseOfx(self.file.file)
   elif self.file.filename.lower().endswith('.qif'):
    items = QifParser.parseQif(self.file.file)
   elif self.file.filename.lower().endswith('.csv'):
    items = QifParser.parseEpaymentsCSV(self.file.file)
   else:
    html += "<strong>Fatal: unrecognised file extension (supported: .qif .ofx)</strong>"
    return html
  except Exception as e:
   html += "<strong>&Eacute;chec&nbsp;: %s</strong></p></div>" % (str(e))
   return html

  html += """
OK
</p>
<table>
<tr><th>Date</th><th>Libelle</th><th>Lecture</th></tr>
"""
  if len(items) < 1:
   html += "<tr><td colspan=\"3\">Pas d'enregistrement dans le fichier&nbsp;!?</td></tr>"
   html += "</table></div>"
   return html
  for item in items:
   line = item.getCamBankCSV()
   try:
    tab = line.split(';')
   except Exception as e:
    html += "<tr><td colspan=\"2\">%s</td><td><strong>&Eacute;chec split&nbsp;: %s</strong>" \
     "</td></tr>" % (line, str(e))
    continue
   try:
    db.execute('set DATESTYLE to European')
    db.execute('insert into raw_operations (odate,olibelle,odetail,omontant) ' \
     'values (%s,%s,%s,%s)' % (self._parent.symcr(tab[0]),
	 self._parent.symcr(tab[1]), self._parent.symcr(tab[2]),
	 self._parent.symcr(tab[3])))
    db.execute('update last_import_date set last_import=1+to_date(\'%s\',\'DD/MM/YYYY\') where last_import < to_date(\'%s\',\'DD/MM/YYYY\') and \'%s\' ~ pattern' % (tab[0], tab[0], self.file.filename))
   except Exception as e:
    html += "<tr><td>%s</td><td>%s</td><td><strong>&Eacute;chec insert&nbsp;: %s</strong>" \
     "</td></tr>" % (tab[0], line, str(e))
    continue
   html += "<tr><td>%s</td><td>%s</td><td>OK</td></tr>" % (tab[0], tab[1])
  db.commit()
  html += "</table></div>"
  return html

 def show_upload_form(self):
  return '''
<div class="box">
<h2>Importer un fichier d'ecritures %s</h2>
<form action="index.cgi" method="POST" enctype="multipart/form-data">
<table>
<tr><th align="right">Choisir un fichier CSV &agrave; importer&nbsp;:</th>
<td><input type="file" name="upload" /></td></tr>
<tr><td></td><td><input type="image" src="../images/valid.png" style="border: 0"/></td>
</table>
</form>
</div>''' % (str(type(self)))


 def load_operations(self):
  """
  Cette fonction lit et affiche la table raw_operations afin
  que l'utilisateur valide les ecritures a inserer dans la table operations.
  """
  last_date = None
  db = self._parent.data
  lines = db.select_all('select oid,%s,%s,%s,%s ' \
   'from raw_operations where onouveau=true' % (self._parent.symde('odate'),
   self._parent.symde('olibelle'), self._parent.symde('odetail'),
   self._parent.symde('omontant')))
  if len(lines) < 1:
   """
   La table est vide, on n'affiche pas le tableau.
   """
   return ''

  bid_filters={}
  for line in db.select_all('select hints, bid from budgets where hints is not null and bshow=true'):
   bid_filters[re.compile(line[0])] = line[1]
  def best_bid(msg):
   for pat in bid_filters.keys():
    if pat.search(msg):
     return bid_filters[pat]

  html = ""
  html += """
<div class="box">
<h2>Insertions des nouvelles op&eacute;rations en base de donn&eacute;es</h2>
<p>
Verifiez qu'il n'y a pas d'erreur avant de valider l'insertion des ecritures.
</p>
<table>
<form>
<tr><th>Date</th><th>Libelle</th><th>Insertion</th></tr>
"""
  for line in lines:
   # self.oids contient la liste des oid selectionnes pour etre inseres
   if self.oids and str(line[0]) in self.oids:
    # l'oid est selectione, on cherche le meilleur budget et on insere
    try:
     db.execute('set DATESTYLE to European')
     db.execute('insert into operations (oid,odate,olibelle,odetail,omontant)' \
      ' values (%d, \'%s\', %s, %s, %d)' % (line[0], line[1],
      self._parent.symcr(line[2]), self._parent.symcr(line[3]),
	  int(line[4].replace(',','').replace('.',''))))
     bb = best_bid(line[2])
     if bb:
      db.execute('update operations set hint_bid=%d where oid=%d' % (bb, line[0]))
     db.execute('update raw_operations set onouveau=false where oid=%d' % (line[0]))
     html += "<tr><td>%s</td><td>%s</td><td><strong>Insertion reussie</strong>" \
      "</td></tr>" % (line[1], line[2])
    except Exception as e:
     # erreur lors de l'insertion
     html += "<tr><td>%s</td><td>%s</td><td><strong>&Eacute;chec&nbsp;: %s</strong>" \
      "</td></tr>" % (line[1], line[2], str(e))
     continue
   else:
    # l'oid n'est pas selectione, on affiche la checkbox
    html += '<tr><td>%s</td><td>%s</td><td><input type="checkbox" name="import" ' \
     'value="%d" checked="checked"/></td></tr>' % (line[1], line[3], line[0])
  html += '<tr><th colspan="3"><input type="submit" value="Ajouter" /></th></tr>'
  html += '</form>'
  html += '</table>'
  html += '</div>'
  db.commit()
  return html

class EditCategories(HTML):
 def __init__(self, cb):
  HTML.__init__(self, cb)
  self.cats = self._parent.get_categories()

 def get_body(self):
  html = ''
  html += self.show_headers()
  html += """
<p>
Les <strong>cat&eacute;gories</strong> permettent de classer les budgets. Par
exemple, les budgets mensuels, les budgets annuels, les autres budgets. Un
budget est obligatoirement rang&eacute; dans une cat&eacute;gorie. Il faut donc
cr&eacute;er au moins une cat&eacute;gorie avant de pouvoir cr&eacute;er un
budget, quitte &agrave; l'appeler <em>Budgets</em>.
</p>
<p>
Pour pouvoir supprimer compl&egrave;tement une cat&eacute;gorie, il faut
qu'elle ne contienne aucun budget.<!-- Sinon, elle ne sera que masqu&eacute;e.-->
</p>
"""
  html += self.show_editable_categories()
  html += self.show_empty_category()
  html += self.show_retour_accueil()
  html += self.show_footers()
  return html

 def show_editable_categories(self):
  html = ''
  if len(self.cats) > 0:
   html += """
<div class="box">
<h2>&Eacute;diter les cat&eacute;gories</h2>
<table>
<tr><th>Nom</th><th>Ordre d'affichage</th><th></th></tr>
"""
   for cat in self.cats:
    html += """
<tr>
<form>
<td>
<input type="hidden" name="cid" value="%d"/>
<input type="text" name="cnom" value="%s"/>
</td>
<td>
<select name="corder">
%s
</select>
</td>
<td><input type="image" src="../images/valid.png" style="border: 0" title="Valider la modification"/></td>
</form>
<form>
<td>
<input type="hidden" name="delcatid" value="%d"/>
<input type="image" src="../images/delete.png" style="border: 0" title="Supprimer la categorie"/></td>
</form>
</tr>
""" % (cat.cid, cat.cnom, self.get_order_select(cat.corder), cat.cid)
   html += """
</table>
<!--
<p>
<strong>Note</strong>&nbsp;: La suppression d'une cat&eacute;gorie non-vide provoquera sa disparition
de l'&eacute;cran principal mais restera en base de donn&eacute;es.
</p>
-->
</div>
"""
  return html

 def show_empty_category(self):
  html = ''
  html += """
<div class="box">
<h2>Cr&eacute;er une nouvelle cat&eacute;gorie</h2>
<table>
<tr><th>Nom</th><th>Ordre d'affichage</th><th></th></tr>
<tr>
<form>
<td>
<input type="text" name="newcnom" value=""/>
</td>
<td>
<select name="corder">
%s
</select>
</td>
<td><input type="image" src="../images/valid.png" style="border: 0"/></td>
</form>
</tr>
</table>
</div>
""" % (self.get_order_select(10000))
  return html

 def get_order_select(self, corder):
  html = ''
  if corder != 1:
   html += '<option value="1">En premier</option>'
  for cat in self.cats:
   if cat.corder == corder-1:
    continue
   if cat.corder == corder:
    html += '<option selected="selected" value="%d">Pas de modification</option>' \
     % (cat.corder)
    continue
   if cat.corder < corder:
    html += '<option value="%d">Apr&egrave;s %s</option>' % (cat.corder+1, cat.cnom)
   else:
    html += '<option value="%d">Apr&egrave;s %s</option>' % (cat.corder, cat.cnom)
  return html

class EditCategory(HTML):
 def __init__(self, cb, cid):
  HTML.__init__(self, cb)
  self.cat = Category(cb, cid)
  self.buds = self.cat.get_budgets()
  self.cats = cb.get_categories()

 def get_body(self):
  html = ''
  html += self.show_headers()
  html += """
<p>
Cette page permet d'ajouter, de modifier et de supprimer des
<strong>budgets</strong>.
</p>
<p>
Un budget est comparable &agrave; une boite dans laquelle on met un peu
d'argent chaque mois et on peut piocher lorsque le besoin s'en fait sentir.
</p>
<p>
Chaque budget poss&egrave;de un <strong>montant cible</strong> qui indique
quelle somme la boite devrait contenir au final.
</p>
<p>
Un <strong>commentaire</strong> permet de donner des pr&eacute;cisions sur le
fonctionnement du budget. Par exemple, s'agit-il d'un budget mensuel (facture
SFR, loyer...), d'un budget annuel (imp&ocirc;t sur le revenu, assurance
voiture...) ou un budget sans limite (r&eacute;serve au cas o&ugrave;, futur
achat de voiture...)&nbsp;? Combien faut-il placer par mois pour atteindre 400
euros par an&nbsp;? Pour quelle date est-ce que le budget de l'assurance
voiture doit-&ecirc;tre rempli&nbsp;?
</p>
<p>
Un budget se remplit d'<strong>actions</strong> au fil du temps. Un budget qui
contient des actions (donc, qui a un historique) ne peut &ecirc;tre
supprim&eacute;. En revanche, il peut &ecirc;tre d&eacute;sactiv&eacute; et
ainsi disparaitre de la page principale. Un budget d&eacute;sactiv&eacute; peut
&ecirc;tre r&eacute;activ&eacute; &agrave; tout moment &agrave; partir de cette
page.
</p>
"""
  html += self.show_editable_budgets()
  html += self.show_empty_budget()
  html += self.show_retour_accueil()
  html += self.show_footers()
  return html

 def show_editable_budgets(self):
  html = ''
  if len(self.buds) > 0:
   html += """
<div class="box">
<h2>&Eacute;diter les budgets de la cat&eacute;gorie "%s"</h2>
<table>
<tr><th>Nom</th><th>Montant</th><th>Commentaire</th><th>D&eacute;placer vers...</th><th colspan="2"></th></tr>
<th><!-- bouton edit --></th><th><!-- bouton delete --></th></tr>
""" % (self.cat.cnom)
   for bud in self.buds:
    html += """
<tr>
<form>
<input type="hidden" name="bid" value="%d"/>
<td>
<input type="text" name="bnom" value="%s"/>
</td>
<td>
<input type="text" name="bmontant" value="%.2f" size="5"/>
</td>
<td>
<textarea name="bcomment">%s</textarea>
</td>
<td>
<select name="cid">
%s
</select>
</td>
<td><input type="image" src="../images/valid.png" style="border: 0" title="Valider la modification"/></td>
</form>
<form>
<td>
<input type="hidden" name="delbid" value="%d"/>
<input type="image" src="../images/delete.png" style="border: 0" title="Supprimer le budget"/></td>
</form>
</tr>
""" % (bud.bid, bud.bnom, float(bud.bmontant)/100, bud.bcomment or '', self.get_cat_select(bud.cid), bud.bid)
   html += """
</table>
<p>
<strong>Note</strong>&nbsp;: La suppression d'un budget non-vide provoquera sa disparition
de l'&eacute;cran principal mais restera en base de donn&eacute;es.
</p>
</div>
"""
  return html

 def show_empty_budget(self):
  html = ''
  html += """
<div class="box">
<h2>Ajouter un nouveau budget dans "%s"</h2>
<table>
<tr><th>Nom</th><th>Montant cible</th><th>Commentaire</th><th></th></tr>
<tr>
<form>
<input type="hidden" name="cid" value="%d"/>
<td>
<input type="text" name="newbnom" value=""/>
</td>
<td>
<input type="text" name="bmontant" value=""/>
</td>
<td>
<input type="text" name="bcomment" value=""/>
</td>
<td><input type="image" src="../images/valid.png" style="border: 0"/></td>
</form>
</tr>
</table>
</div>
""" % (self.cat.cnom, self.cat.cid)
  return html

 def get_cat_select(self, cid):
  html = ''
  for cat in self.cats:
   if cat.cid == cid:
    html += '<option selected="selected" value="%d">---</option>' \
     % (cat.cid)
   else:
    html += '<option value="%d">%s</option>' % (cat.cid, cat.cnom)
  return html

class Help(HTML):
 def get_body(self):
  html = ''
  html += self.show_headers()
  html += """
<h2>La gestion de compte personnel selon CamBank</h2>
<p>
<ul>
<li>Toutes les entr&eacute;es d'argent servent &agrave; alimenter des budgets.</li>
<li>Chaque sortie d'argent provoque la diminution du budget concern&eacute;.</li>
</ul>
<strong>Exemple</strong> : Chaque mois, je re&ccedil;ois mon salaire et j'alloue tant d'argent pour faire mes courses, tant d'argent pour payer le loyer, tant d'argent de c&ocirc;t&eacute; pour payer les impots en septembre, tant d'argent pour une future voiture, etc. Chaque fois que je paye, que ce soit par carte ou en liquide, la somme est retir&eacute;e au budget en question.
</p>

<p>
L'int&eacute;r&ecirc;t est de VISUALISER les d&eacute;penses.
</p>

<h2>Pour commencer</h2>

<p>
La premi&egrave;re chose &agrave; faire est de cr&eacute;er les budgets. Ils doivent forc&eacute;ment appartenir &agrave; une cat&eacute;gorie (qui ne sert qu'&agrave; organiser).
</p>

<p>
Exemple de cat&eacute;gories:
<ul>
<li>Obligatoire <em>(contiendrait les sommes qui partent tous les mois : EDF, Internet, loyer, remboursement cr&eacute;dit immobilier...)</em></li>
<li>Pr&eacute;visionnel <em>(contiendrait les diff&eacute;rentes sommes &agrave; payer annuellement : impot, assurances...)</em></li>
<li>Fonctionnement <em>(contiendrait les budgets 'Nourriture', 'Habits', 'Sorties'...)</em></li>
</ul>
Ensuite, il faut cr&eacute;er tous les budgets n&eacute;cessaires &agrave; l'organisation de son argent.
</p>

<p>
Chaque budget poss&egrave;de :
<ul>
<li>un <strong>nom</strong> qui permet de l'identifier</li>
<li>une <strong>cat&eacute;gorie</strong> qui permet de l'organiser</li>
<li>un <strong>montant cible</strong> qui permet de visualiser si le budget est plut&ocirc;t plein ou plut&ocirc;t vide <em>(exemple: le budget 'Loyer' fera la taille de mon loyer, il faut le remplir tous les mois de la m&ecirc;me somme. le budget 'impot sur le revenu' fera la taille de l'impot annuel, je peux le remplir tous les mois comme je veux, mais il doit etre rempli pour septembre !)</em></li>
<li>un <strong>commentaire</strong> qui permet de mettre des explications <em>(pour le budget 'impot sur le revenu' je mettrais en commentaire combien il faut placer de c&ocirc;t&eacute; chaque mois)</em></li>
</ul>
</p>

<p>
Toutes ces informations peuvent &ecirc;tre modifi&eacute;es plus tard.
</p>

<p>
&Agrave; propos du <strong>montant cible</strong> : c'est purement indicatif. Un budget peut d&eacute;passer son montant sans probl&egrave;me. Il peut m&ecirc;me &ecirc;tre n&eacute;gatif <em>(oups, j'avais pas pr&eacute;vu assez pour sortir ce mois-ci ! je rattraperai le mois prochain...)</em>
</p>

<p>
Il peut &ecirc;tre int&eacute;ressant d'avoir un budget "argent de c&ocirc;t&eacute;" dans lequel puiser en cas d'impr&eacute;vu, pourquoi pas.
</p>

<h2>La premi&egrave;re op&eacute;ration</h2>

<p>
Je suppose que votre compte bancaire n'est pas vide au moment o&ugrave; vous commencez &agrave; utiliser CamBank.
</p>

<p>
Admettons que vous disposiez de 1234 Euros sur votre compte. Cr&eacute;ez une <em>op&eacute;ration manuelle</em> de 1234 euros avec comme <strong>D&eacute;tails</strong> : "&nbsp;initialisation du compte&nbsp;" par exemple.
</p>

<h2>Affecter une op&eacute;ration &agrave; un budget</h2>

<p>
Les op&eacute;rations non affect&eacute;es apparaissent dans un tableau en haut de la page avant la liste des budgets.
</p>

<p>
Pour chaque op&eacute;ration, il suffit de :
<ul>
<li>choisir un budget</li>
<li>&eacute;ventuellement modifier le montant <em>(pour decouper la somme sur plusieurs budgets)</em></li>
<li>valider</li>
</ul>

<p>
Les entr&eacute;es d'argent comme celle que nous venons de cr&eacute;er et comme les salaires vont forc&eacute;ment &ecirc;tre r&eacute;parties dans diff&eacute;rents budgets. Choisissez un montant &agrave; affecter avant de valider.
</p>

<p>
<strong>Exemple</strong> : 600 euros pour le budget 'loyer', 200 euros pour 'nourriture', etc.
</p>

<p>
Lorsqu'il n'y a plus rien &agrave; allouer, l'op&eacute;ration disparait.
</p>

<h2>Importer son relev&eacute; bancaire</h2>

<p>
Ensuite, t&eacute;l&eacute;chargez sur le site de votre banque un relev&eacute; du compte au format QIF (Money de pr&eacute;f&eacute;rence). On peut ensuite envoyer ce fichier &agrave; CamBank dans la partie "Importer des &eacute;critures".
</p>

<p>
Toutes les op&eacute;rations apparaissent alors, il faut leur allouer un budget :)
</p>

<p>
Personnellement, j'essaie de le faire tous les jours puisque ma banque me permet de prendre un relev&eacute; &agrave; partir de la veille.
</p>

<h2>Historique</h2>

<p>
Chaque budget garde en m&eacute;moire la totalit&eacute; des sommes qui lui ont &eacute;t&eacute; allou&eacute;es ou retir&eacute;es (cliquez sur le nom du budget).
</p>

<h2>L'argent liquide</h2>

<p>
Les op&eacute;rations par ch&egrave;que, carte bleue, pr&eacute;l&egrave;vement ou encore virement bancaire sont automatiquement import&eacute;es depuis le compte bancaire. Mais pour les op&eacute;rations liquides, il y a un peu plus de travail.
</p>

<h3>Les sorties d'argent liquide</h3>

<p>
Lorsque je retire de l'argent au distributeur, l'op&eacute;ration bancaire de retrait apparait dans CamBank mais une seule fois (exemple : j'ai retir&eacute; 100 euros). Lorsque je d&eacute;pense cet argent, je dois affecter une partie de cette op&eacute;ration dans les bons budgets. Ainsi, tant que je n'ai pas d&eacute;pens&eacute; tout mon argent liquide, l'op&eacute;ration de retrait reste dans la liste des op&eacute;rations non affect&eacute;es.
</p>

<h3>Les entr&eacute;es d'argent liquide</h3>

<p>
Il faut les rajouter &agrave; la main gr&acirc;ce au petit formulaire en bas de la page d'accueil.
</p>

<p>
<strong>Exemple</strong> : Je re&ccedil;ois 1000 euros en liquide, je cr&eacute;e donc une op&eacute;ration manuelle de 1000 euros, avec la date et un petit texte pour savoir d'o&ugrave; &ccedil;a vient. Je peux ensuite allouer cet argent &agrave; mes budgets.
</p>

"""
  html += self.show_retour_accueil()
  html += self.show_footers()
  return html

class Upload(UploadImport):
 """
 Upload est utilise lorsqu'un fichier est envoye par le formulaire.
 """
 def __init__(self, cb, file):
  UploadImport.__init__(self, cb)
  self.file = file

class Import(UploadImport):
 """
 Import est utilise pour valider les ecritures a inserer dans la base.
 """
 def __init__(self, cb, oids):
  UploadImport.__init__(self, cb)
  self.oids = oids


class ManageSalary(HTML):
 """
 Permet de definir la liste des actions a affecter au salaire mensuel
 """
 def __init__(self, cb):
  HTML.__init__(self, cb)
  self.cb = cb

 def get_body(self):
  html = ''
  html += self.show_headers()
  html += self.show_manage_salary()
  html += self.show_retour_accueil()
  html += self.show_footers()
  return html

 def show_manage_salary(self):
  blist = "<select name=\"bid\">\n"
  for bud in self.cb.get_budgets():
   if bud.bshow is False:
    continue
   blist += "<option value=\"%d\">%s</option>\n" % \
    (bud.bid, bud.bnom)
  blist += "</select>\n"

  sallist = ''
  saltotal = 0
  for sal in self.cb.get_salary():
   sallist += "<tr><td>%d</td><td>%s</td><td>" % (sal[2]/100, Budget(self.cb, sal[1]).bnom)
   sallist += "<a href=\"index.cgi?salary=1&amp;mid=%d\"><img src=\"../images/delete.png\"" % (sal[0])
   sallist += " title=\"Supprimer cette mensualite\"/></a></td></tr>"
   saltotal += sal[2]

  html = ''
  html += """
<div class="box">
<h2>Ajouter une mensualite</h2>
<form>
<input type="hidden" name="salary" value="1"/>
<table>
<tr><th align="right">Budget :</th><td>%s</td></tr>
<tr><th align="right">Montant :</th><td><input type="text" name="montant"/></td></tr>
<tr><th></th><td><input type="image" src="../images/valid.png" style="border: 0"/></td></tr>
</table>
</form>
</div>
""" % (blist)

  if sallist:
   html += """
<div class="box">
<h2>Gerer les mensualites (%d EUR)</h2>
<table>
<tr><th>Montant</th><th>Budget</th><th>Supprimer</th></tr>
""" % (saltotal/100)
   html += sallist
   html += """
</table>
</div>
"""

  return html

class OperationDetails(HTML):
 """
 Permet de visualiser les differentes actions (et donc budgets) liees a une operation.
 """

 def __init__(self, cb, oid):
  HTML.__init__(self, cb)
  self.op = Operation(cb, oid)
  self.cb = cb

 def get_body(self):
  html = ''
  html += self.show_headers()
  html += self.show_operation_details()
  html += self.show_retour_accueil()
  html += self.show_footers()
  return html

 def show_operation_details(self):
  html = ''
  html += """
<div class="box">
<h2>D&eacute;tails de l'op&eacute;ration %d</h2>
<table>
<tr><th>Date</th><td>%s</td></tr>
<tr><th>Montant</th><td>%.2f</td></tr>
<tr><th>Allou&eacute;</th><td>%.2f</td></tr>
<tr><th>Libell&eacute;</th><td>%s</td></tr>
<tr><th>D&eacute;tails</th><td>%s</td></tr>
</table>
<p>
Liste des affectations a un budget&nbsp;:
</p>
<table>
<tr><th>Date action</th><th>Budget</th><th>Montant</th></tr>
""" % (self.op.oid, strftime('%Y-%m-%d', self.op.odate), float(self.op.omontant)/100, float(self.op.obalance)/100, self.op.olibelle, self.op.odetail)

  buds = {}
  for action in self.op.get_actions():
   if not action.bid in buds:
    try:
     buds[action.bid] = Budget(self.cb, action.bid)
    except:
     html += """
<tr><td>%s</td><td><a href="index.cgi?histobid=%d">%s</a></td><td>%.2f</td></tr>
""" % (strftime('Le %Y-%m-%d &agrave; %H:%M:%S', action.adate), action.bid, 'epargne', float(action.amontant)/100)
     continue
   html += """
<tr><td>%s</td><td><a href="index.cgi?histobid=%d">%s</a></td><td>%.2f</td></tr>
""" % (strftime('Le %Y-%m-%d &agrave; %H:%M:%S', action.adate), action.bid, buds[action.bid].bnom, float(action.amontant)/100)

  html += """
</table>
</div>
"""
  return html

class Preferences(HTML):
 def __init__(self, cb):
  self.cb = cb
  HTML.__init__(self, cb)

 def show_password_box(self):
  return '''
<h2>
Chiffrage des données et mot de passe CamBank
</h2>

<ul>
<li>
Toutes vos données textuelles peuvent être chiffrées avec votre mot de passe
que vous seul connaissez.
</li>
<li>
Les nombres ne sont pas chiffrés. Ceci afin de pouvoir faire des calculs sans
devoir tout décoder à chaque fois.
</li>
<li>
Un piratage de la base de données obtiendrait ainsi une liste d'enregistrements
incompréhensibles avec de vraies valeurs qui seront quasiment impossibles à
utiliser.
</li>
<li>
Cette page permet d'activer ou de désactiver le chiffrement. La clef de
chiffrement est votre mot de passe CamBank.
</li>
<li>
Changer le mot de passe implique un recodage complet de toutes vos données et
peut prendre plusieurs secondes de calcul.
</li>
<li>
En cas d'échec de la procédure, la transaction complète sera annulée et le mot de passe ne sera pas changé.
</li>
<li>
Il n'y a <strong>pas de possibilité de récupérer les données chiffrées si vous perdez le mot de passe</strong>.
</li>
</ul>

<div class="box">
<table>
<form>
<input type="hidden" name="setpass" value="1"/>
<tr>
<th>
<label for="crypt">Chiffrement</label>
</th>
<td>
<input type="radio" name="crypt" value="n" {}>non</input><br/>
<input type="radio" name="crypt" value="y" {}>oui</input>
</td>
</tr>
<tr>
<th>
<label for="curpass">Mot de passe actuel</label>
</th>
<td>
<input type="text" name="curpass" />
</td>
</tr>
<tr>
<th>
<label for="newpass">Nouveau mot de passe</label>
</th>
<td>
<input type="text" name="newpass"/>
</td>
</tr>
<tr>
<th>
Procéder à la modification
</th>
<td>
<input type="submit" value="Go"/>
</td>
</tr>
</form>
</table>
</div>
'''.format(self.cb.sympass is None and 'checked="checked"' or ' ', self.cb.sympass is None and ' ' or 'checked="checked"')

 def get_body(self):
  return \
     self.show_headers() \
   + self.show_password_box() \
   + self.show_retour_accueil() \
   + self.show_footers()

class Epargne(HTML):
 def __init__(self, cb):
  self.cb = cb
  HTML.__init__(self, cb)

 def get_body(self):
  html = ''
  html += self.show_headers()
  html += self.help_epargne()
  html += self.list_epargnes()
  html += self.add_epargne_box()
  html += self.show_retour_accueil()
  html += self.show_footers()
  return html

 def add_epargne_box(self):
  html = """
<div class="box">
<h2>Créer une épargne</h2>
<form>
<input type="hidden" name="epargne" value="1"/>
<table>
<tr>
<th>
Nom de l'épargne :
</th>
<td>
<input type="text" name="nom"/>
</td>
</tr>
<tr>
<th>
Ordre d'allocation :
</th>
<td>
<input type="text" name="order"/>
</td>
</tr>
<tr>
<th>
Minimum annuel à épargner :
</th>
<td>
<input type="text" name="min"/>
</td>
</tr>
<tr>
<th>
Répartition du surplus :
</th>
<td>
<input type="text" name="suppl"/>
</td>
</tr>
<tr>
<th>
Montant maximum de l'épargne :
</th>
<td>
<input type="text" name="max"/>
</td>
</tr>
<tr>
<th>
</th>
<td>
<input type="image" src="../images/valid.png" style="border: 0"/>
</td>
</tr>
</table>
</form>
</div>

"""
  return html

 def help_epargne(self):
  return """
<h3>
Fonctionnement des épargnes
</h3>
<p>
Une fonction permet d'allouer des revenus à l'épargne, soit
«&nbsp;épargner&nbsp;». La fonction d'épargne allouera automatiquement l'argent
en fonction des montants que vous avez saisis dans les cases suivantes. Voici
l'algorithme :
</p>
<ol>
<li>
En début d'année, remise à zéro des jauges des épargnes.
</li>
<li>
Alloue l'argent à chaque épargne, dans l'ordre, jusqu'à ce que la jauge atteigne
la somme «&nbsp;Minimum annuel&nbsp;» et passe au suivant.
</li>
<li>
Une fois que toutes les jauges ont atteint leur minimum annuel, l'argent à
épargner est distribué dans toutes les épargnes dont le montant total n'a pas
dépassé le «&nbsp;Montant maximum&nbsp;». Un pourcentage est calculé en
fonction de la valeur renseignée dans «&nbsp;Surplus annuel&nbsp;» et l'argent
est distribué en fonction.
</li>
</ol>
<p>
Hormis cette fonctionnalité, une épargne fonctionne comme un budget : une
opération peut y attribuer de l'argent et en retirer.
</p>
"""

 def list_epargnes(self):
  html = ""

  if len(self.cb.get_epargnes()) > 0:
   html += """
<div class="box">
<h2>Gérer les épargnes</h2>
<table>
<tr>
<th>Nom</th>
<th>Minimum annuel<br/>à allouer à l'épargne</th>
<th>Répartition<br/>du surplus</th>
<th>Montant maximum<br/>de l'épargne</th>
<th>Balance<br/>(montant total alloué)</th>
<th>Jauge<br/>(montant alloué cette année)</th>
<th>Modifier</th>
<th>Cacher</th>
</tr>
"""
   for ep in self.cb.get_epargnes():
    if ep.eshow is False:
     continue
    html += """
<tr>
<td align="center">%d. %s</td>
<td align="center">%d</td>
<td align="center">%d</td>
<td align="center">%d</td>
<td align="right">%.2f</td>
<td align="right">%.2f</td>
<td align="center"><a href="index.cgi?editeid=%d"><img src="../images/modify.png" title="Modifier cette épargne"/></a></td>
<td align="center"><a href="index.cgi?hideeid=%d"><img src="../images/delete.png" title="Cacher cette épargne"/></a></td>
</tr>
""" % (ep.eorder or 0, ep.ecomment is not None and ep.ename or ep.ename+' ('+ep.ecomment+')',
     ep.emin_jauge/100, ep.esuppl_jauge/100, ep.emax_balance/100, ep.ebalance/100.0, ep.ejauge/100.0, ep.eid, ep.eid)
   html += """
</table>
</div>
"""

  return html


class EditEpargne(Epargne):
 def __init__(self, cb, eid):
  self.cb = cb
  Epargne.__init__(self, cb)
  self.epargne = self.cb.get_epargne(eid)

 def get_body(self):
  html = ''
  html += self.show_headers()
  html += self.help_epargne()
  html += self.edit_epargne_box()
  html += self.show_retour_accueil()
  html += self.show_footers()
  return html

 def edit_epargne_box(self):
  html = """
<div class="box">
<h2>Modifer l'épargne : %s</h2>
<form>
<input type="hidden" name="seteid" value="%d"/>
<table>
<tr>
<th>
Ordre d'allocation :
</th>
<td>
<input type="text" name="order" value="%d"/>
</td>
</tr>
<tr>
<th>
Nom de l'épargne :
</th>
<td>
<input type="text" name="nom" value="%s"/>
</td>
</tr>
<tr>
<th>
Minimum annuel à épargner :
</th>
<td>
<input type="text" name="min" value="%d"/>
</td>
</tr>
<tr>
<th>
Répartition du surplus :
</th>
<td>
<input type="text" name="suppl" value="%d"/>
</td>
</tr>
<tr>
<th>
Montant maximum de l'épargne :
</th>
<td>
<input type="text" name="max" value="%d"/>
</td>
</tr>
<tr>
<th>
</th>
<td>
<input type="image" src="../images/valid.png" style="border: 0"/>
</td>
</tr>
</table>
</form>
</div>

""" % (self.epargne.ename, self.epargne.eid, self.epargne.eorder or 0, self.epargne.ename, self.epargne.emin_jauge/100, self.epargne.esuppl_jauge/100, self.epargne.emax_balance/100)
  return html



# cette fonction prend la liste complete des operations, la trie afin de pouvoir
# filtrer les operations deja traitees et enrichir les operations en attente
# avec la liste des operations similaire
# but : montrer les doublons
def filter_pending_operations(ops):
  ops = sorted(ops, key = lambda x: (x.odate, abs(x.omontant), abs(x.obalance)))
  oplist = []
  curop = None
  while len(ops) > 0:
    newop = ops.pop(0)
    if curop == None:
      if newop.obalance != newop.omontant:
        curop = newop
      continue
    # we have curop and newop
    # --
    # first, if new is same than cur, we enrich cur and skip new
    if curop.odate == newop.odate and \
       curop.omontant == newop.omontant:
      if 'similar' not in curop.__dict__:
        curop.similar = []
      curop.similar.append(newop)
    # and if already processed, we skip it as being cur itself
    if newop.obalance == newop.omontant:
      continue
    oplist.append(curop)
    curop = newop
  if curop is not None:
    oplist.append(curop)
  return oplist


class Op:
  def __init__(self, odate, omontant, obalance):
    self.odate = odate
    self.omontant = omontant
    self.obalance = obalance
  def __repr__(self):
    return '({}, {}, {})'.format(self.odate, self.omontant, self.obalance)

def test_filter_pending_operations():
  h1 = [
    Op(1, 1, 0),
    Op(2, 1, 1),
    Op(3, 1, 0),
    Op(3, 1, 0),
  ]
  h2 = filter_pending_operations(h1)
  assert len(h2) == 3
  assert h2[2] in h2[1].similar

if __name__ == '__main__':
  test_filter_pending_operations()



from distutils.core import setup
setup(name='cambank', author='Camille Huot', author_email='noemail@noemail.com',
  version = '2.3.5', url = 'https://github.com/camillehuot/cambank',
  description='CamBank is a personal finance Python module',
  packages = ['cambank'], package_dir= { 'cambank': 'cambank' }
 )
